! Copyright 1999 Stanford University
!All rights reserved
!
!<
!build_filter
!
!Usage:
!call build_filter (arg1, arg2 arg3)
!
!Input Parameters:
!
!
!Output Parameters:
!
!
!
!Description:
!
!
!
!>
!-------------------------------------------------
!
!Author: Robert Clapp, ESMB 463, 7230253
!
!Date Created:Tue Jun 22 14:11:11 PDT 1999
!
!Purpose:
!
!
module build_3d_filter_mod 
  use radial_mod
  use createnhelixmod
  use lapfac
  integer,private,save  :: n(3),jcmp  !data size
  !sampling of filters in cartesian space
  integer,private,save  :: center(3),na(3)  
  !box size and location of zero lag
  real,private,save     :: o(3),d(3)  !sampling of data
  real,private,save     :: hwidth,amp 
  !width and amplitude of radial smoother
  real,private,save     :: dtheta,dtau !sampling for radial filters
  logical,private,save  :: radial,steer  !whether or no
  contains
!subroutine to initialize or change any of the building parameters
  subroutine init_build_filters(n_in,center_in,na_in,o_in,d_in,jcmp_in&
    ,hwidth_in,amp_in,radial_in,steer_in,dtau_in,dtheta_in)
    integer, optional :: n_in(3),center_in(3),na_in(3),jcmp_in
    real,optional     :: o_in(3),d_in(3),hwidth_in,amp_in,dtau_in,dtheta_in
    logical,optional  :: radial_in,steer_in
    if (present(n_in)) then
      n=n_in
    end if
    if (present(jcmp_in)) then
      jcmp=jcmp_in
    end if
    if (present(center_in)) then
      center=center_in
    end if
    if (present(na_in)) then
      na=na_in
    end if
    if (present(o_in)) then
      o=o_in
    end if
    if (present(d_in)) then
      d=d_in
    end if
    if (present(hwidth_in)) then
      hwidth=hwidth_in
    end if
    if (present(amp_in)) then
      amp=amp_in
    end if
    if (present(radial_in)) then
      radial=radial_in
    end if
    if (present(steer_in)) then
      steer=steer_in
    end if
    if (present(dtau_in)) then
      dtau=dtau_in
    end if
    if (present(dtheta_in)) then
      dtheta=dtheta_in
    end if
  end subroutine 
  subroutine build_2d_filter(filt,rough,ncoef)
    type(nfilter)        :: filt,rough  !filter and rough
    integer              :: ncoef       !number of coeficient
!local variables
    integer              :: n123
    integer,allocatable  :: map(:),nh(:)
    real,allocatable     :: p(:)
    integer              :: i2,i1,iout(2),iref,npts(2),resamp,i
    real                 :: pos(2),theta,tau,reg
    integer              :: itau,itheta
    integer,allocatable  :: found(:,:)
    integer              :: ntheta,ntau,noff
    logical              :: good
!FIRST BUILD FILTER
      call build_radial_space(filt,ntau,ntheta,ncoef)
      call create_radial_smooth(ncoef,ntau,ntheta,rough)
  end subroutine

  subroutine build_radial_space(filt,ntau,ntheta,ncoef)
    integer               :: ntau,ntheta,ncoef
    type(nfilter)         :: filt
    real,allocatable      :: p(:)
    integer,allocatable   :: map(:)
    integer,allocatable   :: itau(:),itheta(:),icmp(:)
    real                  :: otheta,dspace,max1,max2,dist,pos2,five_ang&
      &,v
    integer               :: i1,i2,i
    allocate(p(product(n)),map(product(n)))
    allocate(itau(product(n)),itheta(product(n)),icmp(product(n)))
    call find_p(n,o,d,(/1,1/),p,.true.,90.) !we just found the angles
    if (d(2)<1.) then
      v=8.
    else
      v=8000.
    end if 
    max1=1./d(1)+1.5	
    max2=abs(v/d(2)+1.5)
    five_ang=atan(max2/max1)*45/atan(1.)
    where(p< -five_ang)
      p=-five_ang
end where
    where(p> five_ang)
      p=five_ang
end where
    otheta=minval(p)	
    ntheta=ceiling((maxval(p)-otheta)/dtheta)+1
!THE OTHER DIMENSION IS ACTUALLY A DISTANCE FROM THE ORIGIN THING
    dspace=dtau/d(1)
    max1=n(1)
    max2=max(abs(o(2)),abs(o(2)+d(2)*(n(2)-1)))/d(2)
    ntau=ceiling(sqrt(max1**2+max2**2)/dspace)+1
!loop over
    i=0
    do i3=1,n(3)
    do i2=1,n(2)
      pos2=o(2)/d(2) + i2-1
      do i1=1,n(1)
        i=			i+1
        icmp(i)=ceiling(real(i3)/real(jcmp))
        itheta(i)=nint(real(p(i1+(i2-1)*n(1))-otheta)/dtheta)+1
        dist=sqrt(real(i1**2 + pos2**2))/dspace
        itau(i)=nint(dist)+1
      end do
    end do 
    end do 

    map=itau+(itheta-1)*ntau+(icmp-1)*ntau*ntheta
    deallocate(itau,itheta)
    filt=createnhelix(n(1:3),center,(/0,0,0/),na,map)  
    !create prediction filter
    ncoef=size(filt%hlx(1)%lag)
    deallocate(map)     !record number of coef in f
  end subroutine 



  subroutine create_radial_smooth(ncoef,ntau,ntheta,rough)
    type(nfilter)         :: rough
    integer               :: ntau,ntheta,ncoef
    integer               :: i1,i2
    integer,allocatable   ::  nh(:),map(:)
    integer               ::  npts(3),i2,i,resamp,i1
    npts=(/ntau,ntheta,n(3)/)
    allocate(nh(product(npts)))
    allocate(map(ncoef*product(npts)))

    do i2=1,product(npts)
      do i1=1,ncoef 
        i=i1+(i2-1)*ncoef
        map(i)=i2
      end do
    end do 

    nh=3
    call nallocate(rough,nh,map)
    do i=1,size(rough%hlx)
      rough%hlx(i)%lag(1)=ncoef
      rough%hlx(i)%lag(2)=ncoef*ntau
      rough%hlx(i)%lag(3)=ncoef*ntau*ntheta
      rough%hlx(i)%flt=-.333
    end do
    deallocate(nh,map)
  end subroutine 
end module 
