module wei_obj_D
  use wei_util
  use wei_down
  use wei_run
  use wei_external
  use wei_type
  !---------------
  implicit none
contains
  !--------------------------------------------------------------
  !
  !  Function to initialize a data object either from
  ! 
  !    a file or from then general parameters
  !
  !
  !
  logical function wei_data_init(X,tag)
    type(wedata),       intent(inout) :: X
    character(len=128), intent(in)    :: tag

    wei_data_init=F

    X%tag = tag
    if(isfile(X%tag)) then
       if(.not. data_history_init(X)) then
         write(0,*) "trouble initializing data_history"
         return
       end if
    else
       if(.not. data_param_init(X)) then
         write(0,*) "trouble initializing data_paramater"
         return
       end if
    end if

    wei_data_init=T

  end function wei_data_init

  !----------------------------------------------------------------
  !
  !
  !  Create an auxilary data file
  !
  logical function wei_aux_data_init(X,tag)
    type(wedata),       intent(inout) :: X
    character(len=128), intent(in)    :: tag

    wei_aux_data_init=F

    X%tag = tag
    if(.not. data_scratch_init(X)) then
      write(0,*) "trouble initializing scratch data file"
      return
    end if

    wei_aux_data_init=T

  end function wei_aux_data_init
  !----------------------------------------------------------------
  !
  !  Initialize a tag from history
  !   check to make sure that is dimensions don't disagree
  !    with the general params
  ! 
  logical function data_history_init(X)
    type(wedata), intent(inout) :: X

#ifdef DEBUG
    character(len=128), parameter :: unit='data_history_init'
    if(run%debug) call in(unit)
#endif

    call auxinout(X%tag)
    data_history_init=F
    
    
    X%from = "history"

    call hinaxis(X%tag,X%amx,1); X%amx%l='mx'
    if(.not. coax(X%amx,down%amx)) then
      write(0,*) "Data 1 and amx problem"
      return
    end if
    call hinaxis(X%tag,X%amy,2); X%amy%l='my'
    if(.not. coax(X%amy,down%amy)) then
      write(0,*) "Data 2 and amy problem"
      return
    end if
    call hinaxis(X%tag,X%ahx,3); X%ahx%l='hx'
    if(.not. coax(X%ahx,down%ahx)) then
      write(0,*) "Data 3 and ahx problem"
      return
    end if
    call hinaxis(X%tag,X%ahy,4); X%ahy%l='hy'
    if(.not. coax(X%ahy,down%ahy)) then
      write(0,*) "Data 4 and ahy problem"
      return
    end if
    call hinaxis(X%tag,X%aw ,5); X%aw%l ='w_'
    if(.not. coax(X%aw,down%aw)) then
      write(0,*) "Data 5 and w_ problem"
      return
    end if



    if(.not. data_setup(X,.false.)) then
      write(0,*) "trouble setting up data"
      return
    end if

    data_history_init=T

#ifdef DEBUG
    if(run%debug) call data_report(X)
    if(run%debug) call out(unit)
#endif

    
  end function data_history_init

  !----------------------------------------------------------------
  !
  !
  !
  logical function create_data_file(X,tag)
    type(wedata), intent(in) :: X
    character(len=*)         :: tag
       create_data_file=F 
       call auxscr(tag)
       call to_history("esize",8,tag)
       call houaxes(tag,X%amx,X%amy,X%ahx,X%ahy,X%aw)
       create_data_file=T 
   end function 


  !----------------------------------------------------------------
  !
  !
  !
  logical function data_param_init(X)
    type(wedata), intent(inout) :: X
#ifdef DEBUG
    character(len=128), parameter :: unit='data_param_init' 
    if(run%debug) call in(unit)
#endif

    data_param_init=F
    X%from = "parameter"
    X%amx = down%amx; X%amy = down%amy; 
    X%ahx = down%ahx; X%ahy = down%ahy; 
    X%aw  = down%aw;  
    if(.not. create_data_file(X,X%tag)) then
      write(0,*) "trouble creating data file"
      return
    end if
    if(.not. data_setup(X,.true.)) then
      write(0,*) "trouble setting up data"
      return
    end if

    data_param_init=T
#ifdef DEBUG
    if(run%debug) call data_report(X)
    if(run%debug) call out(unit)
#endif
  end function


  subroutine data_assign(X,val)
   type(wedata) :: X
   complex :: val
   logical ::a

    a=file_val_assign(X%tag,val)
  end subroutine
  !----------------------------------------------------------------
  !
  ! data_scratch_init
  !
  !  Create a data scratch file
  !
  !
  logical function data_scratch_init(X)
    type(wedata), intent(inout) :: X
#ifdef DEBUG
    character(len=128), parameter :: unit='data_scratch_init' 
    if(run%debug) call in(unit)
#endif

    data_scratch_init=F
    X%from = "parameter";X%amx = down%amx; X%amy = down%amy; 
    X%ahx = down%ahx; X%ahy = down%ahy; X%aw  = down%aw;  
    if(.not. create_data_file(X,X%tag)) then
      write(0,*) "trouble  creating data_file"
      return
    end if
    if(.not. data_setup(X,.true.)) then
      write(0,*) "trouble setting up data"
      return
    end if
    data_scratch_init=T

#ifdef DEBUG
    if(run%debug) call data_report(X)
    if(run%debug) call out(unit)
#endif
  end function data_scratch_init
  !----------------------------------------------------------------
  ! 
  ! 
  ! 
  !  The core  routine to setup data
  ! 
  ! 
  logical function data_setup(X,zero)
    type(wedata), intent(inout) :: X
    logical :: zero
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='data_setup' 
    if(run%debug) call in(unit)
#endif
    data_setup=F

    X%bites = 8* product( (/X%amx%n,X%amy%n,X%ahx%n,X%ahy%n,down%nws/) )



    !! FREQUENCY blocking
    if( (1.0*X%aw_local%n)/down%nws /= X%aw_local%n/down%nws ) then
       call data_report(X)
       call seperr('D.aw.n is not a multiple of nws')
    end if
    down%nwb = down%nw_local !/ down%nws




    !allocate array
    if(.not. run%mem_save) then
      allocate(X%bin(X%amx%n,X%amy%n,X%ahx%n,X%ahy%n,down%nws),stat=er)
      if(er/=0) call seperr('cant allocate D.bin')
      X%bin=(0.0,0.0)
    else
      X%bin=>down%block
      X%bin=(0.0,0.0)
    end if


    if(zero .and. run%m_nth==1) then
      if(.not. file_val_assign(X%tag,cmplx(0.,0.))) then
        write(0,*) "trouble setting ",trim(X%tag_0)
        return
      end if
    end if
   if(run%mpi) then
     if(.not.  mpi_data_setup(X,zero)) then
       write(0,*) "trouble initializng data mpi"
       return
     end if
   else
     X%aw_local=X%aw
   end if


   data_setup=T
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function data_setup

  !----------------------------------------------------------------
  !
  !   Setup data structure for local files if
  !     we are running MPI
  !
  !

  logical function mpi_data_setup(X,zero)
    type(wedata), intent(inout) :: X
    integer ::n
    logical :: zero
    character(len=128) :: temp
    mpi_data_setup=F

    if(X%tag_0=="/") then
        temp=""
    else
       temp=trim(run%local_path)
    end if
    X%tag_0=X%tag
    write(X%tag,"(a,a,a)") trim(temp),trim(X%tag_0),run%tag_num
    X%aw_local%n=size(down%iw_first)
    n=X%aw%n
    X%aw%n=size(down%iw_first)
    if(.not. create_data_file(X,X%tag)) then
      write(0,*) "trouble setting up mpi data"
      return
    end if
    X%aw%n=n

    if(.not. zero) then 
         call distribute_data(X)
    else
!       if(.not. file_val_assign(X%tag,cmplx(0.,0.))) then
!          write(0,*) "trouble setting ",trim(X%tag_0)
!          return
!       end if
    end if

    mpi_data_setup=T
  end function

  !----------------------------------------------------------------
  !
  ! Information on the data
  !
  !
  subroutine data_report(X)
    type(wedata), intent(in) :: X
    write(0,*)'        ************************************************'
    write(0,*)'        *       ',trim(X%tag),' from ',trim(X%from)
    write(0,'(a,i3,f14.2,f10.3)       ') "         *    Mx: ",X%amx%n, X%amx%o, X%amx%d
    write(0,'(a,i3,f14.2,f10.3)       ') "         *    My: ",X%amy%n, X%amy%o, X%amy%d
    write(0,'(a,i3,f14.2,f10.3)       ') "         *    Hx: ",X%ahx%n, X%ahx%o, X%ahx%d
    write(0,'(a,i3,f14.2,f10.3)       ') "         *    Hy: ",X%ahy%n, X%ahy%o, X%ahy%d
    write(0,'(a,i3,a,i3,a,f10.2,f10.3,a,i3,a)') "         *     W: ",X%aw%n,&
    "(",X%aw_local%n,")",  X%aw%o,  X%aw%d," ->",down%nwb," blocks"
    if(associated(X%bin)) then
       write(0,'(a,f9.2,a,f9.2,a)        ') "         * block: ",X%bites/1024.," / ",8*size(X%bin)/1024.," Kbytes"
    end if
    write(0,*)'        ************************************************'
  end subroutine data_report
  !--------------------------------------------------------------
  ! 
  ! Read the data 
  ! 
  !
  logical function data_read(X,k)
    type(wedata), intent(inout) :: X
    integer, intent(in) :: k
    integer :: stat
#ifdef DEBUG
    character(len=128), parameter :: unit='data_read'
    if(run%debug) call in(unit)
#endif

    stat=sseek_block(X%tag,k-1,X%bites,0)
    if(X%bites/= sreed(X%tag, X%bin, X%bites)) then
      write(0,*) "trouble reading block",k,run%m_ith
      return
    end if

    data_read=T

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function data_read
  !----------------------------------------------------------------
  logical function data_write(X,k)
    type(wedata), intent(inout) :: X
    integer, intent(in) :: k
    integer :: stat
#ifdef DEBUG
    character(len=128), parameter :: unit='data_write' 
    if(run%debug) call in(unit)
#endif
    data_write=F 

    stat=sseek_block(X%tag,k-1,X%bites,0)
    if(X%bites/= srite(X%tag, X%bin, X%bites)) then
      write(0,*) "trouble writing data block",k
      return
    end if

     call to_history("ifreq",k,X%tag)

    data_write=T 
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function data_write
  !----------------------------------------------------------------
  !
  ! 
  !
  !
  subroutine distribute_data(X)
    type(wedata), intent(inout) :: X
    integer, external :: sseek
    integer :: ierr

#ifdef SEP_MPI
    if(run%mpi) then
       if(run%m_ith==0) ierr= sseek(trim(x%tag_0),0,0)
       ierr=sseek(trim(X%tag),0,0)
       if(0/=mpi_sep_tag_distribute(0,X%tag_0,run%m_ith,X%tag,X%bites/4,&
       down%nwp, down%iw_thread)) &
       call seperr("trouble distributing data ");
    end if
#endif
  end subroutine distribute_data
  !----------------------------------------------------------------
  !
  ! 
  !
  !
  subroutine combine_data(X)
    type(wedata), intent(inout) :: X
    integer :: ierr,i
    integer, external :: sseek
#ifdef SEP_MPI
     i=0;if(down%add) i=1
    if( .not. down%adj .and. run%mpi) then
       if(run%m_ith==0) ierr= sseek(trim(x%tag_0),0,0)
       ierr=sseek(trim(X%tag),0,0)

       if(0/=mpi_sep_tag_combine(run%m_ith,X%tag,0,X%tag_0, &
       X%bites/4,down%nwp,down%iw_thread,i)) &
       call seperr("trouble combining data ");
    end if
    call aux_unlink(trim(X%tag))
#endif
  end subroutine combine_data

  !----------------------------------------------------------------
  !
  !  Copy data
  !
  !
  subroutine data_copy(X,Y)
    type(wedata), intent(inout) :: X
    type(wedata), intent(inout) :: Y
    integer :: iwb
    logical :: res
#ifdef DEBUG
    character(len=128), parameter :: unit='data_copy'
    if(run%debug) call in(unit)
#endif


    do iwb=1,down%nwb
       res= data_read (X,iwb)
       Y%bin = X%bin
       res= data_write(Y,iwb)
    end do

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine data_copy


  !----------------------------------------------------------------
  ! 
  !  Add to data structures
  ! 
  ! 
  subroutine data_add(X,Y)
    type(wedata), intent(inout) :: X
    type(wedata), intent(inout) :: Y
    integer :: iwb
    logical :: res
#ifdef DEBUG
    character(len=128), parameter :: unit='data_add'
    if(run%debug) call in(unit)
#endif

    do iwb=1,down%nwb
       res= data_read (X,iwb)
       res= data_read (Y,iwb)
       Y%bin = Y%bin + X%bin
       res= data_write(Y,iwb)
    end do

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine data_add
  !----------------------------------------------------------------  
end module wei_obj_D
