module wei_down
  use sep
  use wei_var
  use wei_run
  use wei_down_type
  integer,private,save :: idone,iblocks,nblocks,nzb,nwb !Number of blocks for this thread
  !----------------------------------------------------------------



contains
  !----------------------------------------------------------------

  !INITIALIZATION FUNCTION FOR DOWNWARD CONTINUATION
  logical function down_init()
#ifdef DEBUG
    character(len=128), parameter :: unit='process_init'
    call in(unit)
#endif

    down_init=F
    !OPERATION SETUP
    call from_param('adj', down%adj, .true.)
    call from_param('add', down%add, .false.)

    !! inversion begins with adjoint, too
    if(run%operation(1:10) .eq. 'inversion.') then
       down%adj=.true. 
    end if

    !HANDLE CONVERTED WAVE 
    ierr=getch("S_S","d",ierr)+getch("S_G","d",ierr)
    if(2==ierr) then
        down%VS='S_S'   !!       slowness
        down%VG='S_G'   !!       slowness
        down%conv_wave=.true.
    else if( 1==ierr) then
       write(0,*) "Must provide S_S and S_G"
       return
    else
        down%VS='S'
        down%VG='S'
        down%conv_wave=.false.
    end if


    !HANDLE SHARED IMAGE FILE
    if(run%share_image) then
       down%image_lock="LOCK."//trim(run%mach)
       call mpi_sep_close_lock(down%image_lock,-1)
    end if
    

    down%dS='dS' !! delta slowness
    down%D='D'   !!       data
    down%U='U'   !!       downward continued wavefield
    down%R='R'   !!       image
    down%dR='dR' !! delta image
    down%W='W'   !!       wavefield


    !NUMBER OF STEPS IN Z AND W TO HANDLE AT A TIME
    call from_param("nzs",down%nzs,1); down%nzb=0;
    call from_param("nws",down%nws,1); down%nwb=0;



    !!FIGURE OUT VARIABLE SLOWNESS
    call from_param("order",down%order,"ANY")
    if(.not. wei_var_init()) then
      write(0,*) "trouble initializing variable step"
     end if
    if(down%order(1:3) == "VAR") then
    else if(down%order(1:3) == "ANY") then
       down%order="OZ";if(down%nzs > 10) down%order="OF"
    else if(down%order(1:2) == "OF" .or. down%order(1:2) == "OZ" ) then
     else 
     write(0,*) "invalid order ",trim(down%order), " must be OF,OZ,VAR, or ANY"
     return
    end if

    if(run%mem_save) down%order="OF"
  
    !! read in axes info from the parameter file
    if(.not. process_parameters()) then
      write(0,*) "trouble initializing parameters"
      return
    end if


    !IF RUNNING IN MEMORY SAVE WE ARE GOING TO REUSE MEMORY BLOCK
    if(run%mem_save) then
      if(down%nws >1  .or. down%nzs>1) then
        write(0,*) "memory save option only available with nws=1 and nzs=1"
        return
      end if
      allocate(down%block(down%amx%n,down%amy%n,down%ahx%n,down%ahy%n,1))
    end if


    !! number of reference velocities
    call from_param("nref",down%nref,1)

    !! CIG in offset or offset ray parameter
    call from_param("pCIG",down%pCIG,.false.)

    if(run%verb) call down_report()

    down_init=T
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function down_init
  !----------------------------------------------------------------

  !READ IN PARAMETERS FROM PAR FILE
  logical function process_parameters()

    down%amx%l='mx'; call pinaxis(down%amx)
    down%amy%l='my'; call pinaxis(down%amy)

    down%ahx%l='hx'; call pinaxis(down%ahx)
    down%ahy%l='hy'; call pinaxis(down%ahy)

    down%apx%l='px'; call pinaxis(down%apx)
    down%apy%l='py'; call pinaxis(down%apy)

    down%aw%l ='w_'; call pinaxis(down%aw )
    down%az%l ='z_'; call pinaxis(down%az )

    down%bz = down%az; down%bz%n=down%nzs
    down%bw = down%aw; down%bw%n=down%nws

      process_parameters=T
  end function process_parameters
  !----------------------------------------------------------------
  !RESET DOWN VARIABLES
  subroutine down_reset(adj,add,mod_,dat_)
    logical,            intent(in) :: adj,add
    character(len=128), intent(in) :: mod_,dat_

    down%adj  = adj
    down%add  = add
    if(run%restart) down%add=.true.
    down%mod_ = mod_
    down%dat_ = dat_

  end subroutine down_reset
  !----------------------------------------------------------------  
  !PRINT STATUS 
  subroutine down_report()
    call weiline()
!    write(run%io_unit,*)                    "        *       model=",down%mod_
!    write(run%io_unit,*)                    "        *        data=",down%dat_
    call weiline()
    write(run%io_unit,*)                    "        *        pCIG=",down%pCIG
    write(run%io_unit,'(a,i5)')             "         *         nzs=",down%nzs
    write(run%io_unit,'(a,i5)')             "         *         nws=",down%nws
    write(run%io_unit,'(a,i5)')             "         *         nzb=",down%nzb
    write(run%io_unit,'(a,i5)')             "         *         nwb=",down%nwb
    write(run%io_unit,'(a,i5)')             "         *         nref=",down%nref
    call weiline()
    write(run%io_unit,'(a,i4,f13.5,f13.5)') "         *          mx= ",down%amx%n,  down%amx%o,  down%amx%d
    write(run%io_unit,'(a,i4,f13.5,f13.5)') "         *          my= ",down%amy%n,  down%amy%o,  down%amy%d
    write(run%io_unit,'(a,i4,f13.5,f13.5)') "         *          hx= ",down%ahx%n,  down%ahx%o,  down%ahx%d
    write(run%io_unit,'(a,i4,f13.5,f13.5)') "         *          hy= ",down%ahy%n,  down%ahy%o,  down%ahy%d
    write(run%io_unit,'(a,i4,f13.5,f13.5)') "         *          px= ",down%apx%n,  down%apx%o,  down%apx%d
    write(run%io_unit,'(a,i4,f13.5,f13.5)') "         *          py= ",down%apy%n,  down%apy%o,  down%apy%d
    call weiline()
  end subroutine down_report

  !SET THE NUMBER OF BLOCKS WE WILL BE USING
  subroutine set_blocks(n_wb,n_zb,n_blocks)
   integer,optional :: n_zb,n_wb,n_blocks
   if(present(n_zb)) nzb=n_zb
   if(present(n_wb)) nwb=n_wb
   if(present(n_blocks)) nblocks=n_blocks
  end subroutine


  !----------------------------------------------------------------
  !PRINT HOW MUCH WE HAVE FINISHED
  subroutine wei_followup(izb,iwb)
    integer :: iwb,izb
    write(run%io_unit,'( "         *&
     & done ",f5.2,"%","    izb=",i4,"/",i4,"     iwb=",i4,"/",i4)') &
     real(idone)/real(nblocks)*100.,izb,nzb,iwb,nwb
     idone=idone+1
  end subroutine wei_followup
  !----------------------------------------------------------------
  !CONVERT TO RADIANS
  subroutine process_radians()
    down%aw%o = 2*pi * down%aw%o
    down%aw%d = 2*pi * down%aw%d
    down%bw%o = 2*pi * down%bw%o
    down%bw%d = 2*pi * down%bw%d
  end subroutine process_radians

  !MAKE SURE THAT THE NUMBER OF OFFSETS AND P ARE THE SAME
  subroutine check_hp()
    if(down%apx%n /= down%ahx%n) then
       call seperr('apx.n /= ahx.n')
    end if
    if(down%apy%n /= down%ahy%n) then
       call seperr('apy.n /= ahy.n')
    end if
  end subroutine check_hp
  !----------------------------------------------------------------
end module wei_down
