! Information on the  job
! This includes  what job is being run
! Thread number OMP/MPI,etc
module wei_run
  use wei_external
#ifdef SEP_MPI
  use mpi_sep 
#endif
  implicit none


  !we will think of these all as public variable
  type weirun
     character(len=4)   :: tag_num
     character(len=9)   :: mach
     character(len=128) :: operator
     character(len=128) :: operation,file_out,local_path
     logical            :: verb    !! verbose
     logical            :: debug   !! Debug
     logical            :: depar   !! debug in the parallel regions
     logical            :: restart !! whether or not we are restarting the job
     logical            :: omp,mpi !! whether or not this is an omp/mpi job 
     logical            :: share_image !! share image file
     integer            :: io_unit !! unit to write out verb info 
     integer            :: m_ith!! mpi thread number
     integer            :: m_nth !! number of mpi threads
     integer            :: o_nth !! number of omp threads
     logical            :: mem_save !! whether to try to save memory
  end type weirun

   type(weirun) :: run

  contains
!----------------------------------------------------------------
!
!
! Function for initializing the run enviro
!
!

  logical function weirun_init()
    character(len=128) :: operator,operation

#ifdef DEBUG
    character(len=128), parameter :: unit='run_mpi_parallel'
    if(run%debug) call in(unit)
#endif DEBUG
  

    call from_param('verb', run%verb, F)
    call from_param('debug',run%debug,F)
    call from_param('restart',run%restart,F)
    call from_param('depar',run%depar,F)
    if(      run%debug) run%verb =T
    if(.not. run%debug) run%depar=T

    !! whether or not to save memory
    call from_param("mem_save",run%mem_save,.false.)
     


    weirun_init=F

    call from_param('operator',operator,'.....')
    run%operator = operator
    call from_param('operation',operation,'dottest...')
    run%operation = operation

    call from_param("mem_save",run%mem_save,.false.)
    if(run%mem_save .and. run%operation(1:3)/="mig" ) then
      write(0,*) "memory save option only available for migration"
      return
    end if 
      

   call from_param("io_file",run%file_out,"NONE")
   if(run%file_out(1:4)=="NONE") then
     run%io_unit=0
   else 
     run%io_unit=6
     open(run%io_unit,file=run%file_out)
   end if


    !Initialize parallel environment for run
    if(.not. run_parallel()) then
       write(0,*) "trouble initializing the wei_run_parallel"
       return
    end if

   weirun_init=T

   if(run%verb) call run_report()

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end function weirun_init



!----------------------------------------------------------------
!
!  Subroutine for initializing parralel run (omp and mpi)
!
!
!
!
  logical function run_parallel()
#ifdef DEBUG
    character(len=128), parameter :: unit='run_parallel'
    if(run%debug) call in(unit)
#endif
 
    run_parallel=F

    if(.not. run_omp_parallel()) then
      write(0,*) "trouble initializing omp enviro"
      return
    end if
    if(.not. run_mpi_parallel()) then
      write(0,*) "trouble initializing mpi enviro"
      return
    end if
    
    run_parallel=T
#ifdef DEBUG
   if(run%debug) call out(unit)
#endif
  end function run_parallel
    

!----------------------------------------------------------------
!
!
! Subroutine to initialize OMP run environment
!
! Sets number of omp threads
!
!
!
!

  logical function run_omp_parallel()
#ifdef DEBUG
    character(len=128), parameter :: unit='omp_run_parallel'
    if(run%debug) call in(unit)
#endif

#ifdef SEP_OMP
    !$OMP PARALLEL
    run%o_nth = omp_get_num_threads()
    !$OMP END PARALLEL

    run%omp=T
#else
    run%o_nth=1
    run%omp=F
#endif

   run_omp_parallel=.true.



#ifdef DEBUG
   if(run%debug) call out(unit)
#endif
  end function  run_omp_parallel

!----------------------------------------------------------------
!
! Function for setting up mpi parallel enviro for mpi
!
! Turns of verb, debug except for the master thread
! Sets up the local datapath
! Figues out the extra to add for each thread
! records the number of threads and this threads num
!
  logical function run_mpi_parallel()
   character(len=128) :: temp_ch,machine
   integer :: ierr,ilen
#ifdef DEBUG
    character(len=128), parameter :: unit='run_mpi_parallel'
    if(run%debug) call in(unit)
#endif

    run_mpi_parallel=F

#ifdef SEP_MPI
    call MPI_Get_processor_name(machine,ilen,ierr)
    call MPI_COMM_SIZE(MPI_COMM_WORLD, run%m_nth,ierr)
    call MPI_COMM_RANK(MPI_COMM_WORLD, run%m_ith,ierr)
    call from_param("share_image",run%share_image,.false.)
#else
    run%share_image=.false.
    run%m_nth=1
    run%m_ith=0
#endif

    !TURN OFF VERB AND DEBUG FOR ALL BUT ZERO MPI THREAD
    if(run%m_ith /= 0) then
       run%verb=F
       run%debug=F
    end if


#ifdef SEP_MPI
    if(run%m_nth>1) then
       run%local_path="                                             &
       &                                                                       &
       &                    "
       if(0==getch("local_datapath","s",run%local_path)) &
       run%local_path="/net/"//machine(1:ilen)//"/scr1/temp/"

       run%mach=trim(machine(1:8))
       

       if(run%m_ith<10) then
          write(run%tag_num,"(a,i1)") "_",run%m_ith
       else if(run%m_ith <100) then
          write(run%tag_num,"(a,i2)") "_",run%m_ith
       else
          write(run%tag_num,"(a,i3)") "_",run%m_ith
       end if
       write(temp_ch,"(a,a,a,a)")"datapath=",trim(run%local_path),trim(run%tag_num),&
       "DATA_"
       call getch_add_string(temp_ch)
    end if
    run%mpi=T
    if(run%m_nth==1) run%mpi=F
#else
    run%mpi=F
#endif


#ifdef DEBUG
    if(run%debug) call out(unit)
#endif

  run_mpi_parallel=T

  end function run_mpi_parallel


 subroutine run_report()
    write(run%io_unit,*) "operator=",trim(run%operator(1:4))," operation=",trim(run%operation(1:8))
    write(run%io_unit,*) "writing to device ",run%io_unit
    write(run%io_unit,*) "verbose=",run%verb," debug=",run%debug," para debug", run%depar
    write(run%io_unit,*) "restart=",run%restart
    write(run%io_unit,*)  "mpi",run%mpi," mpi_ithread=",run%m_ith," mpi_nthread=",run%m_nth
    write(run%io_unit,*)  "omp",run%omp," omp_nthread=",run%o_nth
 end subroutine


end module wei_run
