!HANDLE FFTS
module wei_fft
  !! FFTW
  !! used for data(mx,my,hx,hy,w)

!IF WE AREN'T USING FFTW FAKE IT WITH SU'S PFFT
#ifndef FFTW
use fftw_sep_fake_mod
#endif


  use wei_down
  use wei_run
  use wei_flags
  use wei_attr
  !---------------
  implicit none
   integer, parameter :: max_p=200
  integer(kind=dbl), private,save :: planF(:,:) !! forward FFT
  integer(kind=dbl), private,save :: planI(:,:) !! inverse FFT
  integer, private,save ::  ndim_p(:,:,:)
  real,      private :: scale
  logical,private,save :: fftmeasure
  integer, pointer :: nplans(:),cur_plan(:)

  !! temp  arrays
  complex, dimension(:,:), allocatable, private :: mmhht
  complex, dimension(:,:,:),         pointer, private ::   hht

  !! shift arrays
  complex, dimension(:,:),           pointer, private :: sftF
  complex, dimension(:,:),           pointer, private :: sftI

contains
  !----------------------------------------------------------------
  !INITIALIZE THE FFT
  subroutine weifft_init()
    integer, dimension(5) :: ni
    real    :: arg
    integer :: ihx,ihy
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='weifft_init'
    if(run%debug) call in(unit)
#endif

    !WHETHER OR NOT TO CALCULATE THE SPEED OF VARIOUS FFT METHODS
    call from_param("fftmeasure",fftmeasure,.false.)

    !! FFT
    ni = (/ down%amx%n, down%amy%n, down%ahx%n, down%ahy%n, 1 /)
    scale = 1./sqrt(1.*product(ni))

    
    !IF WE AREN'T RUNNING IN MEMORY SAVE ALLOCATE A TEMPORARY SPACE
    if(.not. run%mem_save) then
      allocate(mmhht(ni(1)*ni(2)*ni(3)*ni(4),run%o_nth),stat=er)
      if(er/=0) call seperr('cant allocate mmhht')
    end if

    allocate(planF( max_p,run%o_nth))
    allocate(planI( max_p,run%o_nth))
    allocate(ndim_p(5,max_p,run%o_nth))
    allocate(nplans(run%o_nth))
    allocate(cur_plan(run%o_nth))
    nplans=0
   write(0,*) "PLAN",run%o_nth
    cur_plan=0

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine weifft_init

 !GIVEN A DATASET DIMENSION
 !FIND IF WE HAVE ALREADY CREATED AN FFTW PLAN FOR IT
 !IF NOT CREATE ONE
 integer function find_plan(ni,ith)
   integer  :: i,ni(5),ith
   logical :: found

     i=1
     found=.false.
    do while(i <= nplans(ith) .and. .not. found) 
      if(all(ni == ndim_p(:,i,ith))) then
       found=.true.
      else
       i=i+1
     end if
    end do
     if(.not. found) then
       call add_plan(ni,ith)
       i=cur_plan(ith)
       found=.true.
     end if
     find_plan=i
 end function find_plan

 !ADD A NEW PLAN FOR A NEW DIMENSION SIZE
 subroutine add_plan(ni,ith)
  integer, dimension(5) :: ni,ni2
  integer :: ith
    if(nplans(ith)<max_p) then
      cur_plan(ith)=cur_plan(ith)+1
      nplans(ith)=nplans(ith)+1
    else 
     if(cur_plan(ith)==max_p) then
       cur_plan(ith)=1
     else
       cur_plan(ith)=cur_plan(ith)+1
     end if 
      call fftwnd_f77_destroy_plan(planF(cur_plan(ith),ith))
      call fftwnd_f77_destroy_plan(planI(cur_plan(ith),ith))
    end if

    ni2=(/ni(1:4),1/)
    ndim_p(:,cur_plan(ith),ith)=ni2

    if(fftmeasure) then
       !! init  forward plan
       call fftwnd_f77_create_plan(planF(cur_plan(ith),ith),5,ni2,fFTW_FORWARD  &
       ,    fFTW_MEASURE + fFTW_THREADSAFE + fFTW_IN_PLACE)
       !! init backward plan
       call fftwnd_f77_create_plan(planI(cur_plan(ith),ith),5,ni2,fFTW_BACKWARD &
       ,    fFTW_MEASURE + fFTW_THREADSAFE + fFTW_IN_PLACE) 
    else
       !! init  forward plan
       call fftwnd_f77_create_plan(planF(cur_plan(ith),ith),5,ni2,fFTW_FORWARD  &
       ,    fFTW_ESTIMATE + fFTW_THREADSAFE + fFTW_IN_PLACE)
       !! init backward plan
       call fftwnd_f77_create_plan(planI(cur_plan(ith),ith),5,ni2,fFTW_BACKWARD &
       ,    fFTW_ESTIMATE + fFTW_THREADSAFE + fFTW_IN_PLACE)
    end if
  end subroutine add_plan 



  !----------------------------------------------------------------
  !! forward FFT
  subroutine weifft(iws,izs,ith,data)
    integer, intent(in) :: iws,ith,izs
    complex, dimension(:,:,:,:,:), pointer :: data
    integer :: imx,imy
    integer :: iplan
    integer :: nd(4)
#ifdef DEBUG
    character(len=128), parameter :: unit='weifft'
    if(run%depar) call in(unit)
#endif

  

     nd=(/size(data,1),size(data,2), var%ihxb_end(izs) -var%ihxb_beg(izs)+1, &
     var%ihyb_end(izs) -var%ihyb_beg(izs)+1/)
     iplan=find_plan((/nd(1),nd(2),nd(3),nd(4),1/),ith)

    !IF WE DOING THINGS WITH MEMORY SAVING OPTIONS REDUCE THE SIZE
    if(.not. run%mem_save) then
      mmhht(1:product(nd(1:4)),ith) =  reshape( &
      data(:,:,var%ihxb_beg(izs):var%ihxb_end(izs), &
        var%ihyb_beg(izs):var%ihyb_end(izs),iws),(/product(nd(1:4))/))

       call fftwnd_f77_one(planF(iplan,ith),mmhht(:,ith),0)

       data(:,:,var%ihxb_beg(izs):var%ihxb_end(izs),var%ihyb_beg(izs):var%ihyb_end(izs),iws)=&
        reshape(mmhht(1:product(nd(1:4)),ith),nd(1:4))* 1./sqrt(1.*product(nd(1:4)))
     else
       call fftwnd_f77_one(planF(iplan,ith),data(:,:,:,:,ith),0)
       data(:,:,var%ihxb_beg(izs):var%ihxb_end(izs), &
        var%ihyb_beg(izs):var%ihyb_end(izs),iws)=data(:,:,var%ihxb_beg(izs):var%ihxb_end(izs), &
        var%ihyb_beg(izs):var%ihyb_end(izs),iws) *1./sqrt(1.*product(nd(1:4)))
     end if


#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine weifft
  !----------------------------------------------------------------
  !! inverse FFT
  subroutine weiift(iws,izs,ith,data)
    integer, intent(in) :: iws, ith,izs
    complex, dimension(:,:,:,:,:), pointer :: data
    integer :: imx,imy,iplan,nd(4)
#ifdef DEBUG
    character(len=128), parameter :: unit='weiift'
    if(run%depar) call in(unit)
#endif

     nd=(/size(data,1),size(data,2), var%ihxb_end(izs) -var%ihxb_beg(izs)+1, &
     var%ihyb_end(izs) -var%ihyb_beg(izs)+1/)

      iplan=find_plan((/nd(1),nd(2),nd(3),nd(4),1/),ith)

    !! FFT
    if(.not. run%mem_save) then
       mmhht(1:product(nd(1:4)),ith) =  reshape( &
        data(:,:,var%ihxb_beg(izs):var%ihxb_end(izs),&
         var%ihyb_beg(izs):var%ihyb_end(izs),iws),(/product(nd(1:4))/))


        call fftwnd_f77_one(planI(iplan,ith),mmhht(:,ith),0) !! Xform in place


       data(:,:,var%ihxb_beg(izs):var%ihxb_end(izs),var%ihyb_beg(izs):var%ihyb_end(izs),iws)=&
        reshape(mmhht(1:product(nd(1:4)),ith),nd(1:4))* 1./sqrt(1.*product(nd(1:4)))
    else
       call fftwnd_f77_one(planI(iplan,ith),data(:,:,:,:,ith),0) !! Xform in place
       data(:,:,var%ihxb_beg(izs):var%ihxb_end(izs), &
        var%ihyb_beg(izs):var%ihyb_end(izs),iws)=data(:,:,var%ihxb_beg(izs):var%ihxb_end(izs), &
        var%ihyb_beg(izs):var%ihyb_end(izs),iws) *1./sqrt(1.*product(nd))
    end if

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine weiift
  !----------------------------------------------------------------
  !! forward FFT
  subroutine weifft4d(data)
    complex, dimension(:,:,:,:), pointer :: data
#ifdef DEBUG
    character(len=128), parameter :: unit='weifft4d'
    if(run%depar) call in(unit)
#endif

    !! FFT
      call fftwnd_f77_one(planF(1),data,0)
      data =data * 1./sqrt(1.*size(data))

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine weifft4d
  !----------------------------------------------------------------
  !! inverse FFT
  subroutine weiift4d(data)
    complex, dimension(:,:,:,:), pointer :: data
#ifdef DEBUG
    character(len=128), parameter :: unit='weiift4d'
    if(run%depar) call in(unit)
#endif

    !! IFT
    call fftwnd_f77_one(planI(1,1),data,0) !! Xform in place
    data(:,:,:,:)=data * 1./sqrt(1.*size(data))

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine weiift4d
  !----------------------------------------------------------------
end module wei_fft
