module wei_kxmig
  !! fk-fx migration
  !! receive wfld at (izs,iws,ith)
  !! return  wfld downward/upward continued
  !! needs FKop, FXop

  use wei_util
  use wei_run
  use wei_external
  use wei_down
  use wei_obj_S
  use wei_fft
  use wei_kmap
  !---------------
  implicit none

  !! temporary interpolation arrays
  complex, dimension(:,:,:,:,:), pointer, private :: tmp1,tmp2,tmp3
  complex, dimension(:,:), pointer, private :: tmps
  character(len=128) :: twave

contains
  !----------------------------------------------------------------
  !----------------------------------------------------------------
  subroutine weikxmig(wfld,iws,izs,ith,KXMIGop,FKop,FXop)
    complex, dimension(:,:,:,:,:), pointer :: wfld
    integer,external,optional ::    FKop,FXop
    integer :: izs,iws,ith
    integer :: st
#ifdef DEBUG
    character(len=128), parameter :: unit='weikxmig'
#endif
    interface !!-------------------------
       integer function KXMIGop(wfld,iws,izs,ith,FKop,FXop) result(st)
         complex, dimension(:,:,:,:,:), pointer :: wfld
         integer, intent(in)       :: iws,izs,ith
         integer,external,optional :: FKop,FXop
       end function KXMIGop
    end interface !!---------------------
#ifdef DEBUG
    if(run%depar) call in(unit)
#endif

    st = KXMIGop(wfld,iws,izs,ith,FKop,FXop)

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine weikxmig
  !----------------------------------------------------------------

!!$
!!$=head1 NAME
!!$
!!$ weikxmigN_init - Multi-velocity KX-migration initialization
!!$
!!$
!!$=head1 SYNOPSIS
!!$
!!$ ierr= weikxmigN_init()
!!$
!!$
!!$=head1 DESCRIPTION
!!$
!!$ Initialize a multi-reference kx-migration velocity operation
!!$
!!$
!!$=head1 SEE ALSO
!!$
!!$L<weikxmigN> L<weikxmig1_init>
!!$
!!$
!!$=head1 sepweif90
!!$
!!$ sepweif90
!!$
!!$
!!$=cut
!!$
  !----------------------------------------------------------------
  !! N reference slownesses
  integer function weikxmigN_init() result(st)
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='weikxmigN_init'
    if(run%debug) call in(unit)
#endif

      !INITIALIZE TEMPORARY SPACE FOR WAVEFIELDS 
      call init_wave_temps()
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    st=OK
  end function weikxmigN_init
!!$
!!$=head1 NAME
!!$
!!$ weikxmigN - Multi- velocity KX-migration 
!!$
!!$
!!$=head1 SYNOPSIS
!!$
!!$ ierr= weikxmigN(wfld,iws,izs,ith,FKOP,FXOP)
!!$
!!$=head1 PARAMETERS
!!$
!!$=over 4
!!$
!!$=item wfld complex(:,:,:,:,:)
!!$
!!$  Wavefield
!!$
!!$=item iws - integer
!!$
!!$ Current frequency slice
!!$
!!$=item izs
!!$
!!$ Current depth slice
!!$
!!$=item ith
!!$
!!$ Current thread
!!$
!!$=item FKOP FK interface L<weiwem>
!!$
!!$=item FXOP FX interface L<weissf>
!!$
!!$
!!$=back
!!$
!!$=head1 DESCRIPTION
!!$
!!$ Perform  a single-reference kx-migration velocity operation
!!$
!!$          st   = FXop(iws,izs,ifk,ith,wfld) !! f-x op
!!$
!!$          call weiift(iws,izs, izs,ith,wfld)
!!$
!!$          st   = FKop(iws,izs,ifk,ith,wfld) !! f-k op
!!$
!!$          call weifft(iws,izs,ith,wfld)
!!$
!!$
!!$=head1 SEE ALSO
!!$
!!$L<weikxmig1> L<weikxmigN_init>
!!$
!!$
!!$=head1 LIBRARY
!!$
!!$ sepweif90
!!$
!!$
!!$=cut
!!$
  !----------------------------------------------------------------
  !! N reference slownesses
  integer function weikxmigN(wfld,iws,izs,ith,FKop,FXop) result(st)
    complex, dimension(:,:,:,:,:), pointer :: wfld
    integer, intent(in)       :: iws,izs,ith
    integer                   :: ifk,isource,ierr
#ifdef DEBUG
    character(len=128), parameter :: unit='weikxmigN'
#endif
    interface !!----------------------------------------------
       integer function FKop(iws,izs,ifk,ith,wfld) result(st)
         integer, intent(in) :: iws,izs,ifk,ith
         complex, dimension(:,:,:,:,:), pointer :: wfld
       end function FKop
       integer function FXop(iws,izs,ifk,ith,wfld) result(st)
         integer, intent(in) :: iws,izs,ifk,ith
         complex, dimension(:,:,:,:,:), pointer :: wfld
       end function FXop
    end interface !!------------------------------------------
#ifdef DEBUG
    if(run%depar) call in(unit)
#endif
    if(down%adj) then !! MIGRATION {
       !----------------------------------------------------------------       

       !REDO THE KMAPS IN CASE WE HAVE CHANGED THE OFFSET RANGE
       call redo_kmaps(var%ihxb_beg(izs),var%ihxb_end(izs), &
       var%ihyb_beg(izs),var%ihyb_end(izs))


       !LOOP OVER SOURCES
       do isource=1,2 !{
          !SELECT THE SOURCE
          call sellect_slow_comp(isource==1,ith)
          call store_wavefield(wfld,ith,iws) !STORE THE CURRENT WAVEFIELD
          do ifk=1,slo%th(ith)%nrs(izs) !{
             call grab_part_wvfld(wfld,ifk,ith,iws,izs) !GRAB A PORTION OF THE WFLD
             !          if(run%m_ith==2 .and. ith==1) write(0,*) "CHECK 0",izs,iws,isource,sum(cabs(wfld(:,:,:,:,iws)))
             st   = FXop(iws,izs,ifk,ith,wfld) !! f-x op
             !          if(run%m_ith==2 .and. ith==1) write(0,*) "CHECK 1",izs,iws,isource,sum(cabs(wfld(:,:,:,:,iws)))
             call weiift(iws,izs,    ith,wfld)
             !          if(run%m_ith==2 .and. ith==1) write(0,*) "CHECK 2",izs,iws,isource,sum(cabs(wfld(:,:,:,:,iws)))
             st   = FKop(iws,izs,ifk,ith,wfld) !! f-k op
             !          if(run%m_ith==2 .and. ith==1) write(0,*) "CHECK 3",izs,iws,isource,sum(cabs(wfld(:,:,:,:,iws)))
             call weifft(iws,izs,    ith,wfld)
             !          if(run%m_ith==2 .and. ith==1) write(0,*) "CHECK 4",izs,iws,isource,sum(cabs(wfld(:,:,:,:,iws))),sum(cabs(tmp1(:,:,:,:,iws)))
             call add_wavefield(wfld,ith,iws)
             !          if(run%m_ith==2 .and. ith==1) write(0,*) "CHECK 5",izs,iws,isource,sum(cabs(wfld(:,:,:,:,iws))),sum(cabs(tmp1(:,:,:,:,iws)))
          end do !}
          call grab_wavefield(wfld,iws,ith,.true.) !GRAB THE ENTIRE WAVEFIELD
          !          if(run%m_ith==2 .and. ith==1) write(0,*) "CHECK 6",izs,iws,isource,sum(cabs(wfld(:,:,:,:,iws)))
       end do !}

       !----------------------------------------------------------------
       !}
    else             !! MODELLING {
       !----------------------------------------------------------------

       do isource=1,2 !{
          call sellect_slow_comp(isource==1,ith)
          call store_wavefield(wfld,iws,ith)

          do ifk=1,slo%th(ith)%nrs(izs) !{
             call grab_wavefield(wfld,iws,ith,.false.)
             call weiift(iws,izs,    ith,wfld)
             st   = FKop(iws,izs,ifk,ith,wfld) !! f-k op
             call weifft(iws,izs,    ith,wfld)
             st   = FXop(iws,izs,ifk,ith,wfld) !! f-x op

             call store_part_wavefield(wfld,ifk,ith,iws,izs) 
          end do !}
          call grab_wavefield(wfld,iws,ith,.true.)
          !----------------------------------------------------------------
       end do !}
    end if!}

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
    st=OK
  end function weikxmigN
  !----------------------------------------------------------------  


!! multiple fk ops
!!
!!       D(z)
!!       |
!!       v
!!       T1
!!
!!    D1 D2 D3 ...
!!
!!    S1 S2 S3 ...
!!     \ | /
!!       T2
!!       |
!!       v
!!       D(z+dz)







!---------------------------------------------
!
!
!  Initialize storage for wavefields
!
!
subroutine init_wave_temps() 
 integer ::er
if(.not. run%mem_save) then
   !! allocate tmp arrays for interpolation of multiple fk ops
   allocate(tmp1(down%amx%n,down%amy%n,down%ahx%n,down%ahy%n,run%o_nth),stat=er)
   if(er/=0) call seperr('cant allocate tmp1')
   allocate(tmp2(down%amx%n,down%amy%n,down%ahx%n,down%ahy%n,run%o_nth),stat=er)
   if(er/=0) call seperr('cant allocate tmp2')
!GET RID OF
   allocate(tmp3(down%amx%n,down%amy%n,down%ahx%n,down%ahy%n,run%o_nth),stat=er)
   if(er/=0) call seperr('cant allocate tmp2')
else 
   allocate(tmps(down%amx%n,down%amy%n))
   write(twave,"(a,a,a)") trim(run%local_path),"T_WAVE",run%tag_num
   call auxtmp(twave)
end if
end subroutine

!-------------------------------------------------
!
!
! Store the wavefield
!
!
subroutine store_w(wfld)
 complex :: wfld(:,:,:,:,:)
 integer :: ierr
  ierr= sseek(twave,0,0)
  ierr= srite(twave,wfld(:,:,:,:,1),8*size(wfld(:,:,:,:,1)))
end subroutine

subroutine store_wavefield(wfld,ith,iws)
 complex :: wfld(:,:,:,:,:)
 integer :: ierr,iws,ith,i1,i2
  
if(run%mem_save) then
  ierr= sseek(twave,0,0)
  ierr= srite(twave,wfld(:,:,:,:,iws),8*size(wfld(:,:,:,:,iws)))
  tmps=cmplx(0.,0.)
   do i2=1,size(wfld,4)
     do i1=1,size(wfld,3)
        ierr=srite(twave,tmps,size(tmps)*8)
     end do
   end do
else
  tmp2(:,:,:,:,ith) = wfld(:,:,:,:,iws)
  tmp1(:,:,:,:,ith) = 0.
end if
end subroutine

!------------------------------------------------------
!
!
!
!
subroutine grab_part_wvfld(wfld,ifk,ith,iws,izs)
integer :: ith,iws,ifk,ierr,izs
 complex :: wfld(:,:,:,:,:)

if(run%mem_save) then
  ierr= sseek(twave,0,0)
  ierr= sreed(twave,wfld(:,:,:,:,iws),8*size(wfld(:,:,:,:,iws)))
  call partition_slow(wfld(:,:,:,:,iws),ifk,izs,ith)
else
    if(slo%th(ith)%source) then
     where(slo%si_s(:,:,:,:,izs) == ifk)
       wfld(:,:,:,:,iws) = tmp2(:,:,:,:,ith)
     elsewhere
         wfld(:,:,:,:,iws) = 0.
     end where
    else
     where(slo%si_r(:,:,:,:,izs) == ifk)
       wfld(:,:,:,:,iws) = tmp2(:,:,:,:,ith)
     elsewhere
         wfld(:,:,:,:,iws) = 0.
     end where
    end if
end if
!if(run%m_ith==2 .and. ith==1) write(0,*) "TM2" ,ith,sum(cabs(tmp2(:,:,:,:,ith)))
end subroutine
!------------------------------------------------------
!
!
!

subroutine add_wavefield(wfld,ith,iws)
complex :: wfld(:,:,:,:,:)
integer :: ith,iws,i1,i2,ierr

if(run%mem_save) then
  ierr= sseek(twave,size(wfld)*8,0)
   do i2=1,size(wfld,4)
     do i1=1,size(wfld,3)
       ierr= sreed(twave,tmps,size(tmps)*8)
       wfld(:,:,i1,i2,iws)=wfld(:,:,i1,i2,iws)+tmps
     end do
   end do 
  ierr= sseek(twave,size(wfld)*8,0)
  ierr= srite(twave,wfld(:,:,:,:,iws),8*size(wfld(:,:,:,:,iws)))
else
  tmp1(:,:,:,:,ith) = tmp1(:,:,:,:,ith) +  wfld(:,:,:,:,iws)
end if
!if(run%m_ith==2 ) write(0,*) "TM1 ",ith,sum(cabs(tmp1(:,:,:,:,ith))),sum(cabs(wfld(:,:,:,:,iws)))

end subroutine
subroutine grab_w(wfld)
complex :: wfld(:,:,:,:,:)
logical :: first
integer :: ith,iws,i1,i2,ierr
  ierr= sseek(twave,0,0)
  ierr= sreed(twave,wfld(:,:,:,:,1),8*size(wfld(:,:,:,:,1)))
end subroutine
subroutine grab_wavefield(wfld,iws,ith,first)
complex :: wfld(:,:,:,:,:)
logical :: first
integer :: ith,iws,i1,i2,ierr
if(run%mem_save) then
  if(.not. first) then
     ierr= sseek(twave,0,0)
  else
     ierr= sseek(twave,size(wfld)*8,0)
  end if
  ierr= sreed(twave,wfld(:,:,:,:,iws),8*size(wfld(:,:,:,:,iws)))
else  if(first) then
  wfld(:,:,:,:,iws) = tmp1(:,:,:,:,ith)
else
  wfld(:,:,:,:,iws) = tmp2(:,:,:,:,ith)
end if
!if(run%m_ith==2 .and. ith==1) write(0,*) "BOTH",ith,sum(cabs(tmp1(:,:,:,:,ith))),sum(cabs(tmp2(:,:,:,:,ith)))
end subroutine

subroutine store_part_wavefield(wfld,ifk,ith,iws,izs)
 complex wfld(:,:,:,:,:)
  integer :: ifk,ith,iws,ierr,izs,i1,i2
if(run%mem_save) then
  call partition_slow(wfld(:,:,:,:,iws),ifk,izs,ith)
  ierr= sseek(twave,0,0)
   do i2=1,size(wfld,4)
     do i1=1,size(wfld,3)
       ierr= sreed(twave,tmps,size(tmps)*8)
       wfld(:,:,i1,i2,iws)=wfld(:,:,i1,i2,iws)+tmps
     end do
   end do 
  ierr= sseek(twave,0,0)
  ierr= srite(twave,wfld(:,:,:,:,iws),8*size(wfld(:,:,:,:,iws)))
else
   where(slo%th(ith)%si(:,:,:,:,izs) == ifk)
      tmp1(:,:,:,:,ith) = tmp1(:,:,:,:,ith) + wfld(:,:,:,:,iws)
   end where
end if
!if(run%m_ith==2 .and. ith==1) write(0,*) "TM1 2",ith,sum(cabs(tmp1))
end subroutine
end module
