module wei_obj_W
  use wei_util
  use wei_down
  use wei_run
  use wei_var
  use wei_type
  use wei_obj_D
  !---------------
  implicit none

  !! !!!!!!!!!!!!!!!!!!!!!!!!!!!!
  type(wewfld), public :: W
  !! !!!!!!!!!!!!!!!!!!!!!!!!!!!!

contains
  !----------------------------------------------------------------
  subroutine weibwfld_init(X,tag)
    type(wewfld),       intent(inout) :: X
    character(len=128), intent(in)    :: tag

    select case(run%operator(1:5))
    case('.Rw.D')

       X%tag = tag
       call wavefield_param_init(W)

    case('dSwdR')

       X%tag = tag
       if(isfile(X%tag)) then
          call wavefield_history_init(X)
       else
          call seperr('W /= file')
       end if
    end select

  end subroutine weibwfld_init
  !----------------------------------------------------------------
  subroutine wavefield_history_init(X)
    type(wewfld), intent(inout) :: X
#ifdef DEBUG
    character(len=128), parameter :: unit='wavefield_history_init'
    if(run%debug) call in(unit)
#endif

    call auxinout(X%tag)
    X%from = "history"

    call hinaxis(X%tag,X%amx,1); X%amx%l='mx'
    if(down%amx%n ==0) then
       down%amx = X%amx
    else
       if(.not. coax( X%amx, down%amx)) then
          call wavefield_report(X)
          call seperr('W.amx /= pro.amx')
       end if
    end if

    call hinaxis(X%tag,X%amy,2); X%amy%l='my'
    if(down%amy%n ==0) then
       down%amy = X%amy
    else
       if(.not. coax( X%amy, down%amy)) then
          call wavefield_report(X)
          call seperr('W.amy /= pro.amy')
       end if
    end if

    call hinaxis(X%tag,X%ahx,3); X%ahx%l='hx'
    if(down%ahx%n ==0) then
       down%ahx = X%ahx
    else
       if(.not. coax( X%ahx, down%ahx)) then
          call wavefield_report(X)
          call seperr('W.ahx /= pro.ahx')
       end if
    end if

    call hinaxis(X%tag,X%ahy,4); X%ahy%l='hy'
    if(down%ahy%n ==0) then
       down%ahy = X%ahy
    else
       if(.not. coax( X%ahy, down%ahy)) then
          call wavefield_report(X)
          call seperr('W.ahy /= pro.ahy')
       end if
    end if

    call hinaxis(X%tag,X%aw ,5); X%aw%l ='w_'
    if(down%aw%n ==0) then
       down%aw = X%aw
    else
       if(.not. coax( X%aw, down%aw)) then
          call wavefield_report(X)
          call seperr('W.aw /= pro.aw')
       end if
    end if

    call hinaxis(X%tag,X%az ,6); X%az%l ='z_'
    if(down%az%n ==0) then
       down%az = X%az
    else
       if(.not. coax( X%az, down%az)) then
          call wavefield_report(X)
          call seperr('W.az /= pro.az')
       end if
    end if

    call wavefield_setup(X)
#ifdef DEBUG
    if(run%debug) call wavefield_report(X)
    if(run%debug) call out(unit)
#endif
  end subroutine wavefield_history_init
  !----------------------------------------------------------------
  subroutine wavefield_param_init(X)
    type(wewfld), intent(inout) :: X
#ifdef DEBUG
    character(len=128), parameter :: unit='wavefield_param_init'
    if(run%debug) call in(unit)
#endif

    call auxscr(X%tag)
    X%from = "parameter"

    X%amx = down%amx; call houaxis(X%tag,X%amx,1)
    X%amy = down%amy; call houaxis(X%tag,X%amy,2)
    X%ahx = down%ahx; call houaxis(X%tag,X%ahx,3)
    X%ahy = down%ahy; call houaxis(X%tag,X%ahy,4)
    X%aw  = down%aw;  call houaxis(X%tag,X%aw, 5)
    X%az  = down%az;  call houaxis(X%tag,X%az, 6)

    call to_history("esize",8,X%tag)

    call wavefield_setup(X)
    call wavefield_assign(X,cmplx(0.))
#ifdef DEBUG
    if(run%debug) call wavefield_report(X)
    if(run%debug) call out(unit)
#endif
  end subroutine wavefield_param_init
  !----------------------------------------------------------------
  subroutine wavefield_setup(X)
    type(wewfld), intent(inout) :: X
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='wavefield_setup'
    if(run%debug) call in(unit)
#endif

    !! DEPTH blocking
    if( (1.0*X%az%n)/down%nzs /= X%az%n/down%nzs ) then
       call wavefield_report(W)
       call seperr('W.az.n is not a multiple of W.nzs')
    end if

    !! FREQUENCY blocking
    if( (1.0*X%aw%n)/down%nws /= X%aw%n/down%nws ) then
       call wavefield_report(W)
       call seperr('W.aw.n is not a multiple of W.nws')
    end if

    down%nwb = down%aw%n / down%nws
    down%nzb = down%az%n / down%nzs
    X%bites = 8* product( (/down%amx%n,down%amy%n,down%ahx%n,down%ahy%n/) )

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine wavefield_setup
  !----------------------------------------------------------------
  subroutine wavefield_assign(X,v)
    type(wewfld), intent(inout) :: X
    complex, intent(in) :: v
    integer :: izb, izs
    integer :: iwb, iws
#ifdef DEBUG
    character(len=128), parameter :: unit='wavefield_assign'
    if(run%debug) call in(unit)
#endif

!    bW%bin=v
    do izb=1,down%nzb
       do izs=1,down%nzs
          do iwb=1,down%nwb
             do iws=1,down%nws
!                call wavefield_write(X,bW%bin,izb,izs,iwb,iws)
             end do
          end do
       end do
    end do

#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
  end subroutine wavefield_assign
  !----------------------------------------------------------------  
  subroutine wavefield_report(X)
    type(wewfld), intent(in) :: X

    write(0,*)'        ************************************************'
    write(0,*)'        *       ',trim(X%tag),' from ',trim(X%from)
    write(0,'(a,i3,f10.2,f10.3)       ') "         *    Mx: ",X%amx%n, X%amx%o, X%amx%d
    write(0,'(a,i3,f10.2,f10.3)       ') "         *    My: ",X%amy%n, X%amy%o, X%amy%d
    write(0,'(a,i3,f10.2,f10.3)       ') "         *    Hx: ",X%ahx%n, X%ahx%o, X%ahx%d
    write(0,'(a,i3,f10.2,f10.3)       ') "         *    Hy: ",X%ahy%n, X%ahy%o, X%ahy%d
    write(0,'(a,i3,f10.2,f10.3,a,i3,a)') "         *     W: ",X%aw%n,  X%aw%o,  X%aw%d," ->",down%nwb," blocks"
    write(0,'(a,i3,f10.2,f10.3,a,i3,a)') "         *     Z: ",X%az%n,  X%az%o,  X%az%d," ->",down%nzb," blocks"
    write(0,*)'        ************************************************'
  end subroutine wavefield_report
  !--------------------------------------------------------------
  subroutine wavefield_read(X,data,izb,izs,iwb,iws)
    type(wewfld), intent(in) :: X
    complex, dimension(:,:,:,:,:), pointer :: data
    integer, intent(in) :: izb,izs,iwb,iws
    integer :: k
    integer :: stat
#ifdef DEBUG
    character(len=128), parameter :: unit='wavefield_read'
    if(run%depar) call in(unit)
#endif

    k=(izb-1)*down%nzs*down%nwb*down%nws &
    +         (izs-1)*down%nwb*down%nws &
    +                 (iwb-1)*down%nws &
    +                             iws

    stat=sseek_block(X%tag,k-1,X%bites,0)
    stat= sreed(X%tag, data(:,:,:,:,iws), X%bites)

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine wavefield_read
  !----------------------------------------------------------------
  subroutine wavefield_write(X,data,izb,izs,iwb,iws)
    type(wewfld), intent(in) :: X
    complex, dimension(:,:,:,:,:), pointer :: data
    integer, intent(in) :: izb,izs,iwb,iws
    integer :: k
    integer :: stat
#ifdef DEBUG
    character(len=128), parameter :: unit='wavefield_write'
    if(run%depar) call in(unit)
#endif

    k=(izb-1)*down%nzs*down%nwb*down%nws &
    +         (izs-1)*down%nwb*down%nws &
    +                 (iwb-1)*down%nws &
    +                             iws

    stat=sseek_block(X%tag,k-1,X%bites,0)
    stat= srite(X%tag, data(:,:,:,:,iws), X%bites)

#ifdef DEBUG
    if(run%depar) call out(unit)
#endif
  end subroutine wavefield_write
  !----------------------------------------------------------------
end module wei_obj_W
