module wei_sll
  !! WEI slowness operator
  !! for FK-FX downward continuation
  !! includes salt logic

  use wei_util
  use wei_obj_S
 use lloyd_mod
  !---------------
  implicit none

  real, dimension(:,:,:,:), pointer, private :: sxy
  real, dimension(:), pointer, private :: bound,ref_slow,s_slice

contains
!!$
!!$=head1 NAME
!!$
!!$ weisll_init - Initialize multiple  slowness  lloyd operator
!!$
!!$=head1 SYNOPSIS
!!$
!!$ ierr=weisll_init()
!!$
!!$=head1 DESCRIPTION
!!$
!!$ Initialize multiple slowness operator lloyd
!!$
!!$=head1 SEE ALSO
!!$
!!$L<weisloN_init> L<weislo1> L<weislsN>
!!$
!!$=head1 LIBRARY
!!$
!!$ sepweif90
!!$
!!$=cut
!!$

  !----------------------------------------------------------------
  !! N reference slownesses
  integer function weisll_init() result(st)
    integer :: er
#ifdef DEBUG
    character(len=128), parameter :: unit='weislsN_init'
    if(run%debug) call in(unit)
#endif

    allocate(bound(down%nref),ref_slow(down%nref))
    allocate(s_slice(size(S%s_s%bin,1)*size(S%s_s%bin,2)))
    
   if(.not. run%mem_save) then
      allocate(sxy(down%amx%n,down%amy%n,down%ahx%n,down%ahy%n),stat=er)
      if(er/=0) call seperr('cant allocate sxy')
   end if

    call init_lloyd()


#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    st=OK
  end function weisll_init
!!$
!!$=head1 NAME
!!$
!!$ weisll - Multiple  slowness operator lloyd vel/find
!!$
!!$=head1 SYNOPSIS
!!$
!!$ ierr=weisll()
!!$
!!$=head1 DESCRIPTION
!!$
!!$  multiple slowness operator lloyd
!!$
!!$=head1 SEE ALSO
!!$
!!$L<weisloN_init> L<weislo1> L<weislsN_init>
!!$
!!$=head1 LIBRARY
!!$
!!$ sepweif90
!!$
!!$=cut
!!$
  !----------------------------------------------------------------  
  !! N reference slownesses
  integer function weisll() result(st)
    integer :: izs,ifk,i4,i3,i2,i1
    real    :: ds, so, sz
    real    :: sl,su
    real    :: smin,smax
    integer :: imx, imy
    real    :: ssalt_true
    real    :: ssalt_mean,ssalt_sum
    real    :: slocal
    integer :: ins,ndo,ierr

#ifdef DEBUG
    character(len=128), parameter :: unit='weislsN'
    if(run%debug) call in(unit)
#endif

    !! spread the slowness in its PBA    
    call weislow_push(S)
    call sellect_slow_comp(.true.,1)
    do izs=1,down%nzs
       slo%th(1)%nrs(izs) = down%nref
       if(.not. run%mem_save) sxy(:,:,:,:) =real(S%pba(:,:,:,:,izs))
       s_slice=reshape(S%s_s%bin(:,:,izs),(/size(S%s_s%bin(:,:,izs))/))
       call lloyd_setup(izs) 
    end do
    call sellect_slow_comp(.false.,1)
    do izs=1,down%nzs
       slo%th(1)%nrs(izs) = down%nref
       if(.not. run%mem_save) sxy(:,:,:,:) =aimag(S%pba(:,:,:,:,izs))
       if(down%conv_wave) then
          s_slice=reshape(S%s_g%bin(:,:,izs),(/size(S%s_g%bin(:,:,izs))/))
       else
          s_slice=reshape(S%s_s%bin(:,:,izs),(/size(S%s_s%bin(:,:,izs))/))
       end if
       call lloyd_setup(izs) 

       if(run%verb) then
       end if
    end do
#ifdef DEBUG
    if(run%debug) call out(unit)
#endif
    st=OK
  end function weisll
  !----------------------------------------------------------------
     

subroutine lloyd_setup(izs)
  integer :: i1,i2,i3,i4,izs

      slo%th(1)%nrs(izs)=lloyd_go(s_slice,ref_slow,down%nref,bound,run%m_ith==0)

       slo%th(1)%sz(izs,1:size(ref_slow)) =ref_slow
       slo%th(1)%bound(izs,1:size(bound))= bound


       if(.not. run%mem_save)  then
         do  i4=1,size(sxy,4)
         do  i3=1,size(sxy,3)
         do  i2=1,size(sxy,2)
         do  i1=1,size(sxy,1)
           slo%th(1)%si(i1,i2,i3,i4,izs)=1
           do while(sxy(i1,i2,i3,i4) > bound(slo%th(1)%si(i1,i2,i3,i4,izs)))
               slo%th(1)%si(i1,i2,i3,i4,izs)=slo%th(1)%si(i1,i2,i3,i4,izs)+1
           end do
         end do;end do;end do;end do
       end if


end subroutine

       
end module wei_sll




