module wei_type
  !! WEI type definitions
  use wei_util

  !---------------
  implicit none

  !----------------------------------------------------------------
  type weslow_comp
     character(len=128) :: tag_0  !non-local tag (actual input/output)
     character(len=128) :: tag    
     character(len=128) :: tag_local,tag_var
     character(len=128) :: from
     integer            :: bites,esize
     logical            :: velocity
     type(axis)         :: amx,amy,az

     !! block binary array
     complex, dimension(:,:,:),     pointer :: bin !slowness (mx,my,z)

  end type weslow_comp
  !----------------------------------------------------------------
  type weslow
   type(weslow_comp) :: s_g,s_s
   type(weslow_comp) :: s1,s2
   logical            :: conv_wave 
   complex, dimension(:,:,:,:,:), pointer :: pba !slowness (mx,my,hx,hy,nzs)
   complex, dimension(:,:,:), pointer :: pba_s !slowness (mx,my,nzs)
  end type weslow

  !----------------------------------------------------------------
  type wedata
     character(len=128) :: tag_0
     character(len=128) :: tag
     character(len=128) :: from
     integer            :: bites
     type(axis)         :: amx, amy, ahx, ahy, aw,aw_local

     !! block binary array
     complex, dimension(:,:,:,:,:), pointer :: bin
  end type wedata
  !----------------------------------------------------------------
  !----------------------------------------------------------------  
  type weimag
     character(len=128) :: tag_0
     character(len=128) :: tag,tag_var
     character(len=128) :: from
     integer            :: bites
     integer            :: esize
     type(axis)         :: amx,amy,ahx,ahy,apx,apy,az

     !! block binary array
     complex, dimension(:,:,:,:,:), pointer :: bin
  end type weimag
  !----------------------------------------------------------------
  !----------------------------------------------------------------
  type wewfld
     character(len=128) :: tag
     character(len=128) :: from
     integer            :: bites
     type(axis)         :: amx,amy,ahx,ahy,aw,az
  end type wewfld
  !----------------------------------------------------------------


  !----------------------------------------------------------------
  type wekmap
     real,    dimension(:),         pointer :: kmx,kmy,khx,khy
     real,    dimension(:,:),       pointer :: ksx,ksy,krx,kry !! f(m,h)
  end type wekmap
  !----------------------------------------------------------------
  !----------------------------------------------------------------
  type wexmap
     real,    dimension(:),         pointer :: mx,my,hx,hy
     real,    dimension(:,:),       pointer :: sx,sy,rx,ry
  end type wexmap
  !----------------------------------------------------------------
  type slowmapthread
     integer, dimension(:,:,:,:,:),   pointer :: si      !! f(mx,my,hx,hy,z)
     integer, dimension(:), pointer           :: nrs     !! f(z)
     real,    dimension(:,      :),   pointer :: sz,bound
     logical :: source
  end type
  !----------------------------------------------------------------
  type weslowmap
     !! FKop slowness
     real,    dimension(:,      :),   pointer :: sz_s,sz_r,bound_r,bound_s !! f(z,nfk)
     !! Number of reference slownesses
     integer, dimension(:), pointer           :: nrs_s,nrs_r     !! f(z)
     !! FXop slowness interpolation map
     integer, dimension(:,:,:,:,:),   pointer :: si_r      !! f(mx,my,hx,hy,z)
     integer, dimension(:,:,:,:,:),   pointer :: si_s      !! f(mx,my,hx,hy,z)
     type(slowmapthread),pointer :: th(:)
  end type weslowmap
  !----------------------------------------------------------------



end module wei_type
